from . 			import util 		as 	jvUtil
from .wrapper 	import 					jvWrapper,jvWrapperSuper
from . 			import menu 		as 	jvMenu
from . 			import fs 			as 	jvFs
from .          import event        as  jvEvent



class jv:
	menu    = jvMenu
	wrapper = jvWrapper
	util    = jvUtil
	fs      = jvFs
	event   = jvEvent